import unittest

from cndi.annotations import Component, Autowired
from cndi.binders.message import DefaultMessageBinder, Output, Input
from cndi.binders.message.mqtt import MqttProducerBinding
from cndi.binders.message.utils import MessageChannel
from cndi.env import VARS, loadEnvFromFile, RCN_ENVS_CONFIG
from cndi.initializers import AppInitializer


@Component
class SinkMQTTTest:
    outputChannel1Binding: MessageChannel
    @Output("default-channel-output")
    def setOutputForDefaultChannel(self, messageBinder: MqttProducerBinding):
        self.outputChannel1Binding = messageBinder


@Component
class SinkRabbitMQTTTest:
    outputChannel1Binding: MessageChannel
    @Output("default-channel-output-rabbit")
    def setOutputForDefaultChannel(self, messageBinder: MqttProducerBinding):
        self.outputChannel1Binding = messageBinder


@Input("default-channel-input")
def setInputForDefaultChannel(message):
    print(message)


@Input("default-channel-input-rabbit")
def setInputForDefaultChannel(message):
    print(message)

class MQTTDefaultMessageBinderTest(unittest.TestCase):
    def setUp(self) -> None:
        VARS.clear()
        VARS[f"{RCN_ENVS_CONFIG}.active.profile".lower()] = "mqtt-test"

        loadEnvFromFile("tests/resources/binder_tests.yml")

    def testDefaultMessageBinder(self):
        defaultMessageBinder = DefaultMessageBinder()

    def testWithAppInitializer(self):
        @Autowired()
        def setSink(sink: SinkMQTTTest):
            sink.outputChannel1Binding.send("Hello")

        appInitializer = AppInitializer()
        appInitializer.componentScan("cndi.binders")
        appInitializer.run()

#
# class RabbitMQDefaultMessageBinderTest(unittest.TestCase):
#     def setUp(self) -> None:
#         VARS.clear()
#         VARS[f"{RCN_ENVS_CONFIG}.active.profile".lower()] = "rabbitmq-test"
#
#         loadEnvFromFile("tests/resources/binder_tests.yml")
#
#     def testDefaultMessageBinder(self):
#         defaultMessageBinder = DefaultMessageBinder()
#
#     def testWithAppInitializer(self):
#         @Autowired()
#         def setSink(sink: SinkRabbitMQTTTest):
#             print(sink)
#             sink.outputChannel1Binding.send("Hello")
#
#         appInitializer = AppInitializer()
#         appInitializer.componentScan("cndi.binders")
#         appInitializer.run()


from functools import wraps


class EventsTypes:
    ON_ENV_LOAD="on_env_load"

REGISTERED_EVENTS = list()

class Event(object):
    def __init__(self, eventType, eventCallback):
        self.eventType = eventType
        self.eventCallback = eventCallback



def OnEvent(eventType):
    def inner_function(func):
        annotations = func.__annotations__

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        REGISTERED_EVENTS.append(Event(eventType=eventType, eventCallback=wrapper))

        return wrapper
    return inner_function




import copy
import importlib
import os

from cndi.binders.message import DefaultMessageBinder
import logging

from cndi.annotations import beanStore, workOrder, beans, components, componentStore, autowires, getBeanObject, getBean, \
    validateBean, queryOverideBeanStore, validatedBeans
from cndi.env import loadEnvFromFile, getContextEnvironment, reload_envs
from cndi.exception import BeanNotFoundException
from cndi.flask.flask_app import FlaskApplication
from cndi.http.management import ManagementServer
from cndi.utils import importSubModules
from cndi.version import VERSION

logger = logging.getLogger(__name__)

initializerComponents = [
    '.'.join([ManagementServer.__module__, ManagementServer.__name__]),
    '.'.join([FlaskApplication.__module__, FlaskApplication.__name__])
]

class AppInitializer:
    def __init__(self):
        """
        Responsible to initialise Dependency Injection for Application
        """
        logger.info(f"CNDI Version: v{VERSION}")
        try:
            from dotenv import load_dotenv
            load_dotenv("./.env")
            reload_envs()
        except ImportError as e:
            logger.error(e)
        self.componentsPath = list()
        applicationYml = "resources/application.yml"

        if os.path.exists(applicationYml):
            logger.info(f"External Configuration found: {applicationYml}")
            loadEnvFromFile(applicationYml)


    def componentScan(self, module):
        importModule = importlib.import_module(module)
        self.componentsPath.append(importModule)


    def run(self):
        """
        Performing Dependency Injection, on priority basis
        Steps Involved in DI
            1. Load Modules and Sub Modules for Bean/Component scanning
            2. Create list for the Available Beans and Components
            3. Resolve Dependency Tree for Beans and Components and Sort in reverse tree dependency
            4. For component classes run postConstruct method if available
            5. Read Configuration for binders and initialise binders for given type (i.e rabbitmq, mqtt)
            6. Perform Dependency Injection by calling setter methods
            7. Start Binder Configuration
        :return: None
        """

        for module in self.componentsPath:
            importSubModules(module)

        for bean in beans:
            validBean = validateBean(bean['fullname'])
            if not validBean:
                continue
            else:
                validatedBeans.append(bean)

        workOrderBeans = workOrder(validatedBeans)

        for bean in workOrderBeans:
            logger.info(f"Registering Bean {bean['fullname']}")
            kwargs = dict()
            for key, className in bean['kwargs'].items():
                tempBean = beanStore[className]
                kwargs[key] = copy.deepcopy(tempBean['object']) if tempBean['newInstance'] else tempBean['object']

            functionObject = bean['object']
            fullname = ".".join([functionObject.__module__, functionObject.__qualname__])
            validBean = validateBean(fullname)
            if validBean:
                bean['objectInstance'] = bean['object'](**kwargs)
                beanStore[bean['name']] = bean
            else:
                logger.debug(f"Ignoring Bean {fullname} due to bean not satisfy")

        for component in components:
            validBean = validateBean(component.fullname)
            if not validBean:
                logger.debug(f"Ignoring Component {component.fullname} due to bean not satisfy")
                continue

            componentStore[component.fullname] = component
            logger.info(f"Building Component: {component.fullname}")
            kwargs = constructKeyWordArguments(component.annotations)
            logger.info("Initializing")
            objectInstance = component.func(**kwargs)
            if 'postConstruct' in dir(objectInstance):
                postConstructKArgs = constructKeyWordArguments(objectInstance.postConstruct.__annotations__)
                objectInstance.postConstruct(**postConstructKArgs)

            override = queryOverideBeanStore(component.fullname)
            if override is not None:
                overrideType = override['overrideType']
                component.fullname = ".".join([overrideType.__module__, overrideType.__name__])

            beanStore[component.fullname] = dict(objectInstance=objectInstance,
                                                 name=component.fullname,
                                                 object=objectInstance, index=0, newInstance=False,
                                                 fullname=component.func.__name__, kwargs=kwargs)

        messageBinderEnabled = getContextEnvironment("rcn.binders.message.enable", defaultValue=False, castFunc=bool)
        defaultMessageBinder = None

        if messageBinderEnabled:
            defaultMessageBinder = getBeanObject(".".join([DefaultMessageBinder.__module__, DefaultMessageBinder.__name__]))
            defaultMessageBinder.performInjection()

        for autowire in autowires:
            autowire.dependencyInject()

        if defaultMessageBinder is not None:
            defaultMessageBinder.start()

        for componentName, componentClass in componentStore.items():
            if componentName in initializerComponents:
                objectInstance = getBeanObject(componentName)
                componentClass.func.run(objectInstance)


def constructKeyWordArguments(annotations):
    kwargs = dict()
    for key, classObject in annotations.items():
        beanName = f"{classObject.__module__}.{classObject.__name__}"
        if beanName in beanStore:
            tempBean = beanStore[beanName]
            kwargs[key] = copy.deepcopy(tempBean['object']) if tempBean['newInstance'] else tempBean['object']
        else:
            raise BeanNotFoundException(f"Following Bean failed to load in Context: "+beanName)

    return kwargs



import re
import time
from datetime import timedelta

import logging
from cndi.utils import File


logger = logging.getLogger(__name__)

class S3BaseChannel:
    def __init__(self, bucketName, func, localDir, pollDelay: timedelta,fileFilter=None,
                 deleteAfterPoll=False,**kwargs):
        self.bucketName = bucketName
        self.func = func
        self.pollDelay = pollDelay.total_seconds()
        self.localDir = File(localDir)
        self.deleteAfterPoll = deleteAfterPoll
        self.localDir.mkdir()
        self.fileFilter = fileFilter

    def pullObject(self, key: str):
        pass

    def process(self):
        pass

    def run(self):
        while True:
            self.process()
            time.sleep(self.pollDelay)

class FileFilter:
    def __init__(self, filter):
        self.filter = filter
        self.pattern = re.compile(self.filter)

    def matches(self, objectName):
        return self.pattern.match(objectName) is not None

class MinioS3Channel(S3BaseChannel):
    def __init__(self, url, bucketName, accessKey, accessSecret, func, localDir,
                 deleteAfterPoll=False, filter:FileFilter = None,
                 pollDelay:timedelta = timedelta(seconds=30), **kwargs):
        from minio import Minio
        S3BaseChannel.__init__(self, bucketName=bucketName,deleteAfterPoll=deleteAfterPoll,
                               pollDelay=pollDelay, func=func, localDir=localDir, **kwargs)

        self.client = Minio(
            url,
            access_key=accessKey,
            secret_key=accessSecret
        )
        self.filter = filter


    def pullObject(self, key: str):
        response = self.client.get_object(self.bucketName, key)
        localFile = self.localDir.resolveChildren(key, isFile=True)
        localFile.mkdir()
        with open(localFile.path, "wb") as stream:
            stream.write(response.read())

        return localFile

    def process(self):
        objects = list(map(lambda x: x.object_name, self.client.list_objects(self.bucketName, recursive=True)))
        objects = list(filter(lambda x: self.filter is None or self.filter.matches(x), objects))
        for object in objects:
            file = self.pullObject(object)
            self.func(file)
            if self.deleteAfterPoll:
                self.client.remove_object(self.bucketName, object)